//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
//
//  Logger.cpp
//  This is a part of the mzscript source code.
//
//  Copyright (C) 2000-2003 the mzscript developers.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#include "logger.hpp"

using std::ofstream;
using std::endl;
using std::ios;

Logger::LogPtr g_Logger(new Logger());

void Logger::logl(LPCSTR pszFormat, ...)
{
	if (fout.is_open())
	{
		char  szMessage[MAX_LINE_LENGTH] = { 0 };
		va_list argList;
	    
		va_start(argList, pszFormat);
		StringCchVPrintf(szMessage, MAX_LINE_LENGTH, pszFormat, argList);
		va_end(argList);

		//format timestamp: "[hh:mm:ss] "
		time_t seconds = time(NULL);
		char* pszTime = ctime(&seconds);
		pszTime += 10;
		pszTime[0] = '[';
		strcpy(pszTime+9,"] ");
	    
		fout << pszTime << szMessage << endl;
	}
}

bool Logger::open(const std::string& sFile)
{
	fout.open(sFile.c_str(), ios::app);
	if (fout.is_open())
	{
		time_t seconds = time(NULL);
		char* pszTime = ctime(&seconds);
		pszTime[24] = '\0';

		fout << "<MZScript log opened " << pszTime << ">" << endl;
		return true;
	}
	else
	{
		return false;
	}
}

void Logger::close()
{
	if (fout.is_open())
	{
		time_t seconds = time(NULL);
		char* pszTime = ctime(&seconds);
		pszTime[24] = '\0';

		fout << "<MZScript log closed " << pszTime << ">" << endl;
		fout.close();
	}
}